/*
 * GCat's Master System Emulator
 * Copyright (c) 2012 James Perry
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

/*==============================================================================
 *
 * General machine state, startup, etc.
 *
 *============================================================================*/

typedef struct screenState_s {
    int scanline;            /* 0 */  /* 0-63 = top border, 64-255 = screen 256-312 = bottom border */
    unsigned char *vram;     /* 4 */
    unsigned char *palette;  /* 8 */
    unsigned char regs[16];  /* 12 */
    short vramAddress;       /* 28 */ /* d14 set means accessing palette */
    unsigned char latch;     /* 30 */ /* latched byte */
    unsigned char latchValid; /* 31 */ /* whether there's something in it */
    unsigned char videoStatus; /* 32 */
    unsigned char scanlineCounter;       /* 33 */
    unsigned char lineIntPending;  /* 34 */
    char pad1;                /* 35 */

    unsigned int *colours;    /* 36 */
    unsigned int *paletteConverted; /* 40 */
    unsigned char *tileBitmaps;  /* 44 */

    unsigned int *screenBitmap; /* 48 */
} screenState_t;

typedef struct soundState_s {
    unsigned char channelAVol; /* 0 */
    unsigned char channelBVol; /* 1 */
    unsigned char channelCVol; /* 2 */
    unsigned char channelDVol; /* 3 */
    short channelAFreq;        /* 4 */
    short channelBFreq;        /* 6 */
    short channelCFreq;        /* 8 */
    unsigned char noiseConfig; /* 10 */
    unsigned char selectedChannel; /* 11 */

    unsigned char *soundBuffer; /* 12 */
    int channelACounter;      /* 16 */
    int channelBCounter;      /* 20 */
    int channelCCounter;      /* 24 */

    int noiseCounter; /* 28 */
    unsigned short lfsr;         /* 32 */
    short noiseFreq;             /* 34 */
} soundState_t;

/*
 * This structure holds all the important things the assembly code
 * needs to access. Mostly CPU state
 */
typedef struct z80State_s {
    int tstates; /* 0 */

    unsigned char **readPages; /* 4 */
    unsigned char **writePages; /* 8 */

    unsigned short af; /* 12 */
    unsigned short bc; /* 14 */
    unsigned short de; /* 16 */
    unsigned short hl; /* 18 */

    unsigned short sp; /* 20 */
    unsigned short pc; /* 22 */

    unsigned short ix; /* 24 */
    unsigned short iy; /* 26 */
  
    unsigned short afp; /* 28 */
    unsigned short bcp; /* 30 */
    unsigned short dep; /* 32 */
    unsigned short hlp; /* 34 */

    unsigned char i; /* 36 */
    unsigned char r; /* 37 */
    unsigned char iff1; /* 38 */
    unsigned char iff2; /* 39 */

    unsigned int *itable; /* 40 */
    unsigned char *parityTable; /* 44 */
    unsigned char *vhAddTable; /* 48 */
    unsigned char *vhSubTable; /* 52 */

    unsigned int *cbitable; /* 56 */
    unsigned int *editable; /* 60 */
    unsigned int *dditable; /* 64 */
    unsigned int *ddcbitable; /* 68 */

    unsigned char intPending; /* 72 */
    unsigned char imode;      /* 73 */
    unsigned char isiy;       /* 74 */
    unsigned char romMask;       /* 75 */

    unsigned char pagingRegs[4]; /* 76 */

    screenState_t *screenState;   /* 80 */

    unsigned char joypadPort1;    /* 84 */
    unsigned char joypadPort2;    /* 85 */
    unsigned char nmiPending;     /* 86 */
    unsigned char port3f; /* 87 */

    soundState_t *soundState;     /* 88 */

    unsigned short debugInfo[4];  /* 92 */

    /* 100 */
} z80State_t;

/* The CPU state */
static z80State_t z80State;

static screenState_t screenState;
static soundState_t soundState;

/* Instruction tables are in the assembly modules */
extern unsigned int insn_table[256];
extern unsigned int cbinsn_table[256];
extern unsigned int edinsn_table[256];
extern unsigned int ddinsn_table[256];
extern unsigned int ixcbinsn_table[256];

/* This table contains the parity, sign, zero and all undocumented flags for each byte value
 * Useful for the result of a logical operation.
 */
static unsigned char parityTable[256] = {
    0x44, 0x00, 0x00, 0x04, 0x00, 0x04, 0x04, 0x00, 
    0x08, 0x0c, 0x0c, 0x08, 0x0c, 0x08, 0x08, 0x0c, 
    0x00, 0x04, 0x04, 0x00, 0x04, 0x00, 0x00, 0x04, 
    0x0c, 0x08, 0x08, 0x0c, 0x08, 0x0c, 0x0c, 0x08, 
    0x20, 0x24, 0x24, 0x20, 0x24, 0x20, 0x20, 0x24, 
    0x2c, 0x28, 0x28, 0x2c, 0x28, 0x2c, 0x2c, 0x28, 
    0x24, 0x20, 0x20, 0x24, 0x20, 0x24, 0x24, 0x20, 
    0x28, 0x2c, 0x2c, 0x28, 0x2c, 0x28, 0x28, 0x2c, 
    0x00, 0x04, 0x04, 0x00, 0x04, 0x00, 0x00, 0x04, 
    0x0c, 0x08, 0x08, 0x0c, 0x08, 0x0c, 0x0c, 0x08, 
    0x04, 0x00, 0x00, 0x04, 0x00, 0x04, 0x04, 0x00, 
    0x08, 0x0c, 0x0c, 0x08, 0x0c, 0x08, 0x08, 0x0c, 
    0x24, 0x20, 0x20, 0x24, 0x20, 0x24, 0x24, 0x20, 
    0x28, 0x2c, 0x2c, 0x28, 0x2c, 0x28, 0x28, 0x2c, 
    0x20, 0x24, 0x24, 0x20, 0x24, 0x20, 0x20, 0x24, 
    0x2c, 0x28, 0x28, 0x2c, 0x28, 0x2c, 0x2c, 0x28, 
    0x80, 0x84, 0x84, 0x80, 0x84, 0x80, 0x80, 0x84, 
    0x8c, 0x88, 0x88, 0x8c, 0x88, 0x8c, 0x8c, 0x88, 
    0x84, 0x80, 0x80, 0x84, 0x80, 0x84, 0x84, 0x80, 
    0x88, 0x8c, 0x8c, 0x88, 0x8c, 0x88, 0x88, 0x8c, 
    0xa4, 0xa0, 0xa0, 0xa4, 0xa0, 0xa4, 0xa4, 0xa0, 
    0xa8, 0xac, 0xac, 0xa8, 0xac, 0xa8, 0xa8, 0xac, 
    0xa0, 0xa4, 0xa4, 0xa0, 0xa4, 0xa0, 0xa0, 0xa4, 
    0xac, 0xa8, 0xa8, 0xac, 0xa8, 0xac, 0xac, 0xa8, 
    0x84, 0x80, 0x80, 0x84, 0x80, 0x84, 0x84, 0x80, 
    0x88, 0x8c, 0x8c, 0x88, 0x8c, 0x88, 0x88, 0x8c, 
    0x80, 0x84, 0x84, 0x80, 0x84, 0x80, 0x80, 0x84, 
    0x8c, 0x88, 0x88, 0x8c, 0x88, 0x8c, 0x8c, 0x88, 
    0xa0, 0xa4, 0xa4, 0xa0, 0xa4, 0xa0, 0xa0, 0xa4, 
    0xac, 0xa8, 0xa8, 0xac, 0xa8, 0xac, 0xac, 0xa8, 
    0xa4, 0xa0, 0xa0, 0xa4, 0xa0, 0xa4, 0xa4, 0xa0, 
    0xa8, 0xac, 0xac, 0xa8, 0xac, 0xa8, 0xa8, 0xac
};

/*
 * These tables contain the overflow and half carry flag values for additions and subtractions.
 * They are indexed with:
 *  d5 - result bit 7
 *  d4 - input 1 bit 7
 *  d3 - input 2 bit 7
 *  d2 - result bit 3
 *  d1 - input 1 bit 3
 *  d0 - input 2 bit 3
 */
static unsigned char vhAddTable[64] = {
    0x00, 0x10, 0x10, 0x10, 0x00, 0x00, 0x00, 0x10, 
    0x00, 0x10, 0x10, 0x10, 0x00, 0x00, 0x00, 0x10, 
    0x00, 0x10, 0x10, 0x10, 0x00, 0x00, 0x00, 0x10, 
    0x04, 0x14, 0x14, 0x14, 0x04, 0x04, 0x04, 0x14, 
    0x04, 0x14, 0x14, 0x14, 0x04, 0x04, 0x04, 0x14, 
    0x00, 0x10, 0x10, 0x10, 0x00, 0x00, 0x00, 0x10, 
    0x00, 0x10, 0x10, 0x10, 0x00, 0x00, 0x00, 0x10, 
    0x00, 0x10, 0x10, 0x10, 0x00, 0x00, 0x00, 0x10,   
};

static unsigned char vhSubTable[64] = {
    0x00, 0x00, 0x10, 0x00, 0x10, 0x00, 0x10, 0x10, 
    0x04, 0x04, 0x14, 0x04, 0x14, 0x04, 0x14, 0x14, 
    0x00, 0x00, 0x10, 0x00, 0x10, 0x00, 0x10, 0x10, 
    0x00, 0x00, 0x10, 0x00, 0x10, 0x00, 0x10, 0x10, 
    0x00, 0x00, 0x10, 0x00, 0x10, 0x00, 0x10, 0x10, 
    0x00, 0x00, 0x10, 0x00, 0x10, 0x00, 0x10, 0x10, 
    0x04, 0x04, 0x14, 0x04, 0x14, 0x04, 0x14, 0x14, 
    0x00, 0x00, 0x10, 0x00, 0x10, 0x00, 0x10, 0x10, 
};

/* 
 * normal read pages, then normal write pages, then first ROM pointer,
 * then paging regs pointer, then cartridge RAM start, then internal
 * RAM start
 */
static unsigned char *pageTable[20];

static int cartridgeSize; /* in 16KB pages */

static unsigned char *cartridgeROM;
static unsigned char *cartridgeRAM;
static unsigned char *internalRAM; /* second 8K is for ROM writes */
static unsigned char *videoRAM;

static unsigned char *tileBitmaps;
static unsigned int *screenBitmap;

static unsigned char paletteRAM[32];
static unsigned int paletteConverted[32];
static unsigned int colours[64];


static unsigned char soundBuffer[312];

static void resetCPU()
{
    /* initialise the CPU state */
    z80State.tstates = -3500000;

    z80State.readPages = &pageTable[0];
    z80State.writePages = &pageTable[8];

    z80State.af = 0x0000;
    z80State.bc = 0x0000;
    z80State.de = 0x0000;
    z80State.hl = 0x0000;

    z80State.sp = 0x0000;
    z80State.pc = 0x0000;
    z80State.ix = 0x0000;
    z80State.iy = 0x0000;

    z80State.afp = 0x0000;
    z80State.bcp = 0x0000;
    z80State.dep = 0x0000;
    z80State.hlp = 0x0000;

    z80State.i = 0x00;
    z80State.r = 0x00;
    z80State.iff1 = 0x00;
    z80State.iff2 = 0x00;

    z80State.itable = insn_table;
    z80State.parityTable = parityTable;
    z80State.vhAddTable = vhAddTable;
    z80State.vhSubTable = vhSubTable;

    z80State.cbitable = cbinsn_table;
    z80State.editable = edinsn_table;
    z80State.dditable = ddinsn_table;
    z80State.ddcbitable = ixcbinsn_table;

    z80State.imode = 1;
    z80State.intPending = 0;
    z80State.nmiPending = 0;

    /*
     * FIXME: internationalisation info?
     */
    z80State.joypadPort1 = 0xff;
    z80State.joypadPort2 = 0xff;

    z80State.romMask = (cartridgeSize - 1);

    z80State.debugInfo[0] = 0;
    z80State.debugInfo[1] = 0;
    z80State.debugInfo[2] = 0;
    z80State.debugInfo[3] = 0;

    z80State.pagingRegs[0] = 0;
    z80State.pagingRegs[1] = 0;
    z80State.pagingRegs[2] = 1;
    z80State.pagingRegs[3] = 2;

    pageTable[0] = &cartridgeROM[0x00000];
    pageTable[1] = &cartridgeROM[0x02000];
    pageTable[2] = &cartridgeROM[0x04000];
    pageTable[3] = &cartridgeROM[0x06000];
    pageTable[4] = &cartridgeROM[0x08000];
    pageTable[5] = &cartridgeROM[0x0a000];
    pageTable[6] = &internalRAM[0x00000];
    pageTable[7] = &internalRAM[0x00000];

    pageTable[8] = &internalRAM[0x02000];
    pageTable[9] = &internalRAM[0x02000];
    pageTable[10] = &internalRAM[0x02000];
    pageTable[11] = &internalRAM[0x02000];
    pageTable[12] = &internalRAM[0x02000];
    pageTable[13] = &internalRAM[0x02000];
    pageTable[14] = &internalRAM[0x00000];
    pageTable[15] = &internalRAM[0x00000];

    pageTable[16] = cartridgeROM;
    pageTable[17] = (&z80State.pagingRegs[0] - 0x1ffc);
    pageTable[18] = cartridgeRAM;
    pageTable[19] = internalRAM;

    memset(cartridgeRAM, 0, 32*1024);
    memset(internalRAM, 0, 8*1024);
    memset(videoRAM, 0, 16*1024);
    memset(paletteRAM, 0, 32);
    memset(paletteConverted, 0, 32*sizeof(unsigned int));
    memset(tileBitmaps, 0, 32*1024);

    z80State.screenState = &screenState;
    screenState.scanline = 0;
    screenState.vram = videoRAM;
    screenState.palette = &paletteRAM[0];
    memset(screenState.regs, 0, 16);
    screenState.vramAddress = 0;
    screenState.latch = 0;
    screenState.latchValid = 0;
    screenState.videoStatus = 0;
    screenState.paletteConverted = &paletteConverted[0];
    screenState.colours = &colours[0];
    screenState.tileBitmaps = tileBitmaps;
    screenState.screenBitmap = screenBitmap;
    screenState.lineIntPending = 0;

    z80State.soundState = &soundState;
    memset(&soundState, 0, sizeof(soundState));
    soundState.soundBuffer = &soundBuffer[0];
    soundState.lfsr = 0x8000;
    soundState.channelAVol = 0xf;
    soundState.channelBVol = 0xf;
    soundState.channelCVol = 0xf;
    soundState.channelDVol = 0xf;
}


/*
 * Master System memory map:
 *
 * 0000 - 03FF : always first 1K of ROM
 * 0400 - 3FFF : selectable ROM bank 0
 * 4000 - 7FFF : selectable ROM bank 1
 * 8000 - BFFF : selectable ROM bank 2 or cartridge RAM
 * C000 - DFFF : on-board RAM (8K)
 * E000 - FFFB : mirror of on-board RAM
 * FFFC - FFFF : paging registers
 *
 * Paging registers:
 *
 * FFFC : bit 2 - selects page 0 or 1 of cartridge RAM
 *        bit 3 - 0 means map ROM at 8000, 1 means cartridge RAM
 * FFFD : ROM bank 0 select
 * FFFE : ROM bank 1 select
 * FFFF : ROM bank 2 select
 *
 * A cartridge ROM can be any power of 2 size from 32K to 512K
 */

static void initialiseColours();

/*
 * Initialises the emulator
 *
 * Returns 1 on success, 0 on failure
 */
int MS_startup()
{
    cartridgeSize = 2;
    cartridgeROM = malloc(512 * 1024);     /* max cartridge size */
    cartridgeRAM = malloc(32 * 1024);
    internalRAM = malloc(16 * 1024);
    videoRAM = malloc(16 * 1024);
    tileBitmaps = malloc(32 * 1024);
    screenBitmap = malloc(256 * 192 * sizeof(int));

    if ((!cartridgeROM) || (!cartridgeRAM) || (!internalRAM) ||
	(!videoRAM) || (!tileBitmaps) || (!screenBitmap)) {
	return 0;
    }

    cartridgeROM[0] = 0xf3; /* di */
    cartridgeROM[1] = 0xc3; /* jp 0x0001 */
    cartridgeROM[2] = 0x01;
    cartridgeROM[3] = 0x00;

    cartridgeROM[0x66] = 0xc9; /* ret */

    initialiseColours();
  
    resetCPU();

    return 1;
}

int MS_loadROM(char *filename)
{
    int len;
    unsigned char *s, *romstart;
    FILE *f;

    f = fopen(filename, "rb");
    if (!f) {
	return 0;
    }
    fseek(f, 0, SEEK_END);
    len = ftell(f);
    fseek(f, 0, SEEK_SET);
    
    s = malloc(len);
    if (!f) {
	fclose(f);
	return 0;
    }
    fread(s, 1, len, f);
    fclose(f);

    romstart = s;

    /* deal with weird 512 byte "header" some files have */
    if (len & 512) {
	len -= 512;
	romstart += 512;
    }

    if ((len > (512*1024)) || (len & 0x3fff)) {
	free(s);
	return 0;
    }

    /* copy in ROM data */
    cartridgeSize = len >> 14;
    memcpy(cartridgeROM, romstart, len);

    /* uncomment to get straight to Labyrinth on Sonic 1 */
    /*cartridgeROM[0x1c9f] = 0x3e;
      cartridgeROM[0x1ca0] = 0x0a;
      cartridgeROM[0x1ca1] = 0x32;
      cartridgeROM[0x1ca2] = 0x3e;
      cartridgeROM[0x1ca3] = 0xd2;
      cartridgeROM[0x1ca4] = 0x00;
      cartridgeROM[0x1ca5] = 0x00;*/

    /* reset machine */
    resetCPU();

    free(s);
    return 1;
}

/*
 * Haven't converted the snapshotting routines to RasPi yet.
 */
#if 0
/*
 * Snapshotting
 *
 * would need to save:
 *  - which ROM we're running - a pathname for this. Java app can deal
 *    with this part
 *
 *  - all CPU registers (26 bytes)
 *  - interrupt mode and enabled status (3 bytes)
 *  - internal RAM contents (8K)
 *  - cartridge RAM contents (???) (32K)
 *  - paging registers (4 bytes)
 *
 *  - video RAM contents (16K)
 *  - palette RAM contents (32 bytes)
 *  - video register contents (16 bytes)
 *  - VRAM current address and latched byte (4 bytes)
 *
 *  - sound channel volumes (4 bytes)
 *  - sound channel frequencies (6 bytes)
 *  - noise channel info (1 byte)
 *  - which channel currently selected for writing (1 byte)
 *
 * Total size: 56K + 97 bytes
 */
/*jint
Java_uk_org_gcat_ms_GcatMSApp_saveState( JNIEnv* env,
					 jobject thiz,
					 jbyteArray state,
					 jint idx)*/
int MS_saveState(char *filename)
{
    // FIXME: make this work!
    int len;
    unsigned char *s, *p;

    len = (*env)->GetArrayLength(env, state);
    if (len < ((56*1024)+97)) {
	/* insufficient space */
	return 0;
    }

    p = (unsigned char *)(*env)->GetByteArrayElements(env, state, NULL);

    s = p + idx;

    /* save CPU registers */
    s[0] = z80State.af & 0xff;
    s[1] = (z80State.af >> 8) & 0xff;
    s[2] = z80State.bc & 0xff;
    s[3] = (z80State.bc >> 8) & 0xff;
    s[4] = z80State.de & 0xff;
    s[5] = (z80State.de >> 8) & 0xff;
    s[6] = z80State.hl & 0xff;
    s[7] = (z80State.hl >> 8) & 0xff;
    s[8] = z80State.sp & 0xff;
    s[9] = (z80State.sp >> 8) & 0xff;
    s[10] = z80State.pc & 0xff;
    s[11] = (z80State.pc >> 8) & 0xff;
    s[12] = z80State.ix & 0xff;
    s[13] = (z80State.ix >> 8) & 0xff;
    s[14] = z80State.iy & 0xff;
    s[15] = (z80State.iy >> 8) & 0xff;
    s[16] = z80State.afp & 0xff;
    s[17] = (z80State.afp >> 8) & 0xff;
    s[18] = z80State.bcp & 0xff;
    s[19] = (z80State.bcp >> 8) & 0xff;
    s[20] = z80State.dep & 0xff;
    s[21] = (z80State.dep >> 8) & 0xff;
    s[22] = z80State.hlp & 0xff;
    s[23] = (z80State.hlp >> 8) & 0xff;
    s[24] = z80State.i;
    s[25] = z80State.r;

    /* other misc system state */
    s[26] = z80State.imode;
    s[27] = z80State.iff1;
    s[28] = z80State.iff2;
    s[29] = z80State.pagingRegs[0];
    s[30] = z80State.pagingRegs[1];
    s[31] = z80State.pagingRegs[2];
    s[32] = z80State.pagingRegs[3];
  
    /* video state */
    memcpy(&s[33], screenState.regs, 16);
    s[49] = (screenState.vramAddress & 0xff);
    s[50] = (screenState.vramAddress >> 8) & 0xff;
    s[51] = screenState.latch;
    s[52] = screenState.latchValid;

    /* sound state */
    s[53] = soundState.channelAVol;
    s[54] = soundState.channelBVol;
    s[55] = soundState.channelCVol;
    s[56] = soundState.channelDVol;
    s[57] = soundState.channelAFreq & 0xff;
    s[58] = (soundState.channelAFreq >> 8) & 0xff;
    s[59] = soundState.channelBFreq & 0xff;
    s[60] = (soundState.channelBFreq >> 8) & 0xff;
    s[61] = soundState.channelCFreq & 0xff;
    s[62] = (soundState.channelCFreq >> 8) & 0xff;
    s[63] = soundState.noiseConfig;
    s[64] = soundState.selectedChannel;

    /* save RAMs */
    memcpy(&s[65], internalRAM, 8192);
    memcpy(&s[65 + (8*1024)], cartridgeRAM, 32768);
    memcpy(&s[65 + (40*1024)], videoRAM, 16384);
    memcpy(&s[65 + (56*1024)], paletteRAM, 32);

    (*env)->ReleaseByteArrayElements(env, state, (jbyte *)p, 0);
    return 1;
}

/*jint
Java_uk_org_gcat_ms_GcatMSApp_loadState( JNIEnv* env,
					 jobject thiz,
					 jbyteArray state,
					 jint idx)*/
int MS_loadState(char *filename)
{
    // FIXME: make this work!
    int len;
    unsigned char *s, *p;
    int i;
    unsigned int trow, c;

    len = (*env)->GetArrayLength(env, state);
    if (len < ((56*1024)+97)) {
	/* insufficient space */
	return 0;
    }

    p = (unsigned char *)(*env)->GetByteArrayElements(env, state, NULL);

    s = p + idx;

    /* reset machine to default state */
    resetCPU();

    /* restore CPU registers */
    z80State.af = s[0] | (s[1] << 8);
    z80State.bc = s[2] | (s[3] << 8);
    z80State.de = s[4] | (s[5] << 8);
    z80State.hl = s[6] | (s[7] << 8);
    z80State.sp = s[8] | (s[9] << 8);
    z80State.pc = s[10] | (s[11] << 8);
    z80State.ix = s[12] | (s[13] << 8);
    z80State.iy = s[14] | (s[15] << 8);
    z80State.afp = s[16] | (s[17] << 8);
    z80State.bcp = s[18] | (s[19] << 8);
    z80State.dep = s[20] | (s[21] << 8);
    z80State.hlp = s[22] | (s[23] << 8);
    z80State.i = s[24];
    z80State.r = s[25];

    /* system state */
    z80State.imode = s[26];
    z80State.iff1 = s[27];
    z80State.iff2 = s[28];
    z80State.pagingRegs[0] = s[29];
    z80State.pagingRegs[1] = s[30];
    z80State.pagingRegs[2] = s[31];
    z80State.pagingRegs[3] = s[32];

    /* video state */
    memcpy(screenState.regs, &s[33], 16);
    screenState.vramAddress = s[49] | (s[50] << 8);
    screenState.latch = s[51];
    screenState.latchValid = s[52];

    /* sound state */
    soundState.channelAVol = s[53];
    soundState.channelBVol = s[54];
    soundState.channelCVol = s[55];
    soundState.channelDVol = s[56];
    soundState.channelAFreq = s[57] | (s[58] << 8);
    soundState.channelBFreq = s[59] | (s[60] << 8);
    soundState.channelCFreq = s[61] | (s[62] << 8);
    soundState.noiseConfig = s[63];
    soundState.selectedChannel = s[64];

    /* load RAMs */
    memcpy(internalRAM, &s[65], 8192);
    memcpy(cartridgeRAM, &s[65 + (8*1024)], 32768);
    memcpy(videoRAM, &s[65 + (40*1024)], 16384);
    memcpy(paletteRAM, &s[65 + (56*1024)], 32);

    (*env)->ReleaseByteArrayElements(env, state, (jbyte *)p, 0);

    /* load scanline counter from register 10 */
    screenState.scanlineCounter = screenState.regs[10];

    /* update page tables based on paging registers*/
    pageTable[0] = &cartridgeROM[(z80State.pagingRegs[1] & z80State.romMask) << 14];
    pageTable[1] = &cartridgeROM[((z80State.pagingRegs[1] & z80State.romMask) << 14) + 0x2000];
    pageTable[2] = &cartridgeROM[(z80State.pagingRegs[2] & z80State.romMask) << 14];
    pageTable[3] = &cartridgeROM[((z80State.pagingRegs[2] & z80State.romMask) << 14) + 0x2000];
    if (!(z80State.pagingRegs[0] & 8)) {
	pageTable[4] = &cartridgeROM[(z80State.pagingRegs[3] & z80State.romMask) << 14];
	pageTable[5] = &cartridgeROM[((z80State.pagingRegs[3] & z80State.romMask) << 14) + 0x2000];
	pageTable[12] = &internalRAM[0x02000];
	pageTable[13] = &internalRAM[0x02000];
    }
    else {
	pageTable[4] = &cartridgeRAM[((z80State.pagingRegs[0] & 4) << 12)];
	pageTable[5] = &cartridgeRAM[((z80State.pagingRegs[0] & 4) << 12) + 0x2000];
	pageTable[12] = &cartridgeRAM[((z80State.pagingRegs[0] & 4) << 12)];
	pageTable[13] = &cartridgeRAM[((z80State.pagingRegs[0] & 4) << 12) + 0x2000];
    }

    /* convert palette to our format */
    for (i = 0; i < 32; i++) {
	paletteConverted[i] = colours[paletteRAM[i] & 63];
    }

    /* convert tiles to our format */
    for (i = 0; i < 4096; i++) {
	/* loop over each tile row */
	trow = *(unsigned int *)(&videoRAM[i << 2]);

	c = ((trow >> 7) & 1) | ((trow >> 14) & 2) | ((trow >> 21) & 4) |
	    ((trow >> 28) & 8);
	tileBitmaps[(i << 3) + 0] = c;
	c = ((trow >> 6) & 1) | ((trow >> 13) & 2) | ((trow >> 20) & 4) |
	    ((trow >> 27) & 8);
	tileBitmaps[(i << 3) + 1] = c;
	c = ((trow >> 5) & 1) | ((trow >> 12) & 2) | ((trow >> 19) & 4) |
	    ((trow >> 26) & 8);
	tileBitmaps[(i << 3) + 2] = c;
	c = ((trow >> 4) & 1) | ((trow >> 11) & 2) | ((trow >> 18) & 4) |
	    ((trow >> 25) & 8);
	tileBitmaps[(i << 3) + 3] = c;
	c = ((trow >> 3) & 1) | ((trow >> 10) & 2) | ((trow >> 17) & 4) |
	    ((trow >> 24) & 8);
	tileBitmaps[(i << 3) + 4] = c;
	c = ((trow >> 2) & 1) | ((trow >> 9) & 2) | ((trow >> 16) & 4) |
	    ((trow >> 23) & 8);
	tileBitmaps[(i << 3) + 5] = c;
	c = ((trow >> 1) & 1) | ((trow >> 8) & 2) | ((trow >> 15) & 4) |
	    ((trow >> 22) & 8);
	tileBitmaps[(i << 3) + 6] = c;
	c = ((trow) & 1) | ((trow >> 7) & 2) | ((trow >> 14) & 4) |
	    ((trow >> 21) & 8);
	tileBitmaps[(i << 3) + 7] = c;
    }

    /* set noise frequency properly */
    switch (soundState.noiseConfig & 3) {
    case 0:
	soundState.noiseFreq = 0x10;
	break;
    case 1:
	soundState.noiseFreq = 0x20;
	break;
    case 2:
	soundState.noiseFreq = 0x40;
	break;
    case 3:
	soundState.noiseFreq = soundState.channelBFreq;
	break;
    }
  
    return 1;
}

#endif

void MS_getAudioData(unsigned short *audio)
{
    int i;
    /* convert to 16 bits per sample and invert */
    for (i = 0; i < 312; i++) {
	audio[i] = ((soundBuffer[i] << 8) ^ 0x8000);
    }
}

/* assembly function which runs the Z80 until tstates overflows */
int asmExecute(z80State_t *state, unsigned char *mem);

/*
 * Runs the Z80 CPU for one frame
 */
int MS_execute()
{
    int result;

    z80State.tstates = -224;

    result = asmExecute(&z80State, NULL);

    return (int) z80State.pc;
}

static void initialiseColours()
{
    int i;
    for (i = 0; i < 64; i++) {
	colours[i] = 0xff000000 | ((i & 3) << 6) | ((i & 12) << 12) |
	    ((i & 48) << 18);
    }
}

int MS_getDebugWord(int idx)
{
    return (int) z80State.debugInfo[idx];
}

void MS_triggerNMI()
{
    z80State.nmiPending = 1;
}

void MS_setJoypadPort1(int val)
{
    z80State.joypadPort1 = (val & 0xff);
}

unsigned int MS_getBorderColour()
{
    return paletteConverted[0x10 + (screenState.regs[7] & 0xf)];
}

/*
 * 0 = normal mode (256x192)
 * 1 = left hand column blanked (248x192)
 * 2 = screen disabled
 */
int MS_getScreenMode()
{
    if (!(screenState.regs[1] & 0x40)) {
	/* screen disabled */
	return 2;
    }

    if (screenState.regs[0] & 0x20) {
	/* left column blanked */
	return 1;
    }

    return 0;
}

static void renderLine(int line, unsigned int *dest)
{
    int i;
    int sr, tr, rwt, urwt;
    int hs;
    int tmaddr, staddr;
    unsigned int tile, trow, c;
    unsigned int palor;

    int sx, sy, sh, srow;
    int ns, sxa;

    /* line number goes from 0-191 */

    /* horizontal scroll */
    hs = screenState.regs[8] & 0xff;

    /* scrolled row */
    sr = screenState.regs[9] + line;
    if (sr >= 224) sr -= 224;

    /* handle fixed panel at top */
    if ((screenState.regs[0] & 0x40) && (line < 16)) {
	sr = line;
	hs = 0;
    }

    /* tile row */
    tr = (sr >> 3);

    /* row within tile */
    rwt = (sr & 7);
    urwt = 7 - rwt;  /* for upside-down tiles */

    /* tilemap address */
    tmaddr = (screenState.regs[2] & 14) << 10;
    tmaddr += (tr << 6);

    /* render 32 background tiles */
    for (i = 0; i < 32; i++) {
	/*if (videoRAM[tmaddr+1] & 0x10) {
	  hs += 8;
	  tmaddr += 2;
	  continue;
	  }*/

	tile = videoRAM[tmaddr] | ((videoRAM[tmaddr+1] & 1) << 8);

	/* sprite palette? */
	palor = (videoRAM[tmaddr+1] & 8) << 1;

	/* flip vertical? */
	if (videoRAM[tmaddr+1] & 4) {
	    trow = ((tile << 6) | (urwt << 3));
	}
	else {
	    trow = ((tile << 6) | (rwt << 3));
	}

	/* flip horizontal? */
	if (!(videoRAM[tmaddr+1] & 2)) {
	    c = tileBitmaps[trow];
	    dest[(hs&255)] = paletteConverted[c | palor];
	    c = tileBitmaps[trow+1];
	    dest[(hs+1)&255] = paletteConverted[c | palor];
	    c = tileBitmaps[trow+2];
	    dest[(hs+2)&255] = paletteConverted[c | palor];
	    c = tileBitmaps[trow+3];
	    dest[(hs+3)&255] = paletteConverted[c | palor];
	    c = tileBitmaps[trow+4];
	    dest[(hs+4)&255] = paletteConverted[c | palor];
	    c = tileBitmaps[trow+5];
	    dest[(hs+5)&255] = paletteConverted[c | palor];
	    c = tileBitmaps[trow+6];
	    dest[(hs+6)&255] = paletteConverted[c | palor];
	    c = tileBitmaps[trow+7];
	    dest[(hs+7)&255] = paletteConverted[c | palor];
	}
	else {
	    c = tileBitmaps[trow];
	    dest[(hs+7)&255] = paletteConverted[c | palor];
	    c = tileBitmaps[trow+1];
	    dest[(hs+6)&255] = paletteConverted[c | palor];
	    c = tileBitmaps[trow+2];
	    dest[(hs+5)&255] = paletteConverted[c | palor];
	    c = tileBitmaps[trow+3];
	    dest[(hs+4)&255] = paletteConverted[c | palor];
	    c = tileBitmaps[trow+4];
	    dest[(hs+3)&255] = paletteConverted[c | palor];
	    c = tileBitmaps[trow+5];
	    dest[(hs+2)&255] = paletteConverted[c | palor];
	    c = tileBitmaps[trow+6];
	    dest[(hs+1)&255] = paletteConverted[c | palor];
	    c = tileBitmaps[trow+7];
	    dest[hs&255] = paletteConverted[c | palor];
	}
	hs += 8;
	tmaddr += 2;
    }

    /* render sprites */
    ns = 0;
    staddr = (screenState.regs[5] & 0x7e) << 7;
    sh = 8;
    if (screenState.regs[1] & 3) sh = 16;
    sxa = 0;
    if (screenState.regs[0] & 8) sxa = 8;
    for (i = 0; i < 64; i++) {
	sy = videoRAM[staddr+i];
	if (sy == 208) break;
	if (ns >= 8) break;

	if ((sy <= line) && ((sy + sh) > line)) {
	    sx = videoRAM[staddr+128+(i<<1)] - sxa;
	    tile = videoRAM[staddr+129+(i<<1)];
	    if (screenState.regs[1] & 2) tile &= 0xfe;
	    if (screenState.regs[6] & 4) tile |= 256;

	    srow = line - sy;

	    if (screenState.regs[1] & 1) {
		/* double sized sprite */
		srow >>= 1;
		trow = ((tile << 6) | (srow << 3));
		c = tileBitmaps[trow];
		if (c) {
		    dest[sx&255] = paletteConverted[c | 0x10];
		    dest[(sx+1)&255] = paletteConverted[c | 0x10];
		}
		c = tileBitmaps[trow+1];
		if (c) {
		    dest[(sx+2)&255] = paletteConverted[c | 0x10];
		    dest[(sx+3)&255] = paletteConverted[c | 0x10];
		}
		c = tileBitmaps[trow+2];
		if (c) {
		    dest[(sx+4)&255] = paletteConverted[c | 0x10];
		    dest[(sx+5)&255] = paletteConverted[c | 0x10];
		}
		c = tileBitmaps[trow+3];
		if (c) {
		    dest[(sx+6)&255] = paletteConverted[c | 0x10];
		    dest[(sx+7)&255] = paletteConverted[c | 0x10];
		}
		c = tileBitmaps[trow+4];
		if (c) {
		    dest[(sx+8)&255] = paletteConverted[c | 0x10];
		    dest[(sx+9)&255] = paletteConverted[c | 0x10];
		}
		c = tileBitmaps[trow+5];
		if (c) {
		    dest[(sx+10)&255] = paletteConverted[c | 0x10];
		    dest[(sx+11)&255] = paletteConverted[c | 0x10];
		}
		c = tileBitmaps[trow+6];
		if (c) {
		    dest[(sx+12)&255] = paletteConverted[c | 0x10];
		    dest[(sx+13)&255] = paletteConverted[c | 0x10];
		}
		c = tileBitmaps[trow+7];
		if (c) {
		    dest[(sx+14)&255] = paletteConverted[c | 0x10];
		    dest[(sx+15)&255] = paletteConverted[c | 0x10];
		}
	    }
	    else {
		trow = ((tile << 6) | (srow << 3));
		c = tileBitmaps[trow];
		if (c) dest[sx&255] = paletteConverted[c | 0x10];
		c = tileBitmaps[trow+1];
		if (c) dest[(sx+1)&255] = paletteConverted[c | 0x10];
		c = tileBitmaps[trow+2];
		if (c) dest[(sx+2)&255] = paletteConverted[c | 0x10];
		c = tileBitmaps[trow+3];
		if (c) dest[(sx+3)&255] = paletteConverted[c | 0x10];
		c = tileBitmaps[trow+4];
		if (c) dest[(sx+4)&255] = paletteConverted[c | 0x10];
		c = tileBitmaps[trow+5];
		if (c) dest[(sx+5)&255] = paletteConverted[c | 0x10];
		c = tileBitmaps[trow+6];
		if (c) dest[(sx+6)&255] = paletteConverted[c | 0x10];
		c = tileBitmaps[trow+7];
		if (c) dest[(sx+7)&255] = paletteConverted[c | 0x10];
	    }
	    ns++;
	}
    }


    /* tilemap address */
    tmaddr = (screenState.regs[2] & 14) << 10;
    tmaddr += (tr << 6);

    /* horizontal scroll */
    hs = screenState.regs[8] & 0xff;

    /* render tiles in front of sprites */
    for (i = 0; i < 32; i++) {
	if (!(videoRAM[tmaddr+1] & 0x10)) {
	    /* in front of sprites, done already */
	    hs += 8;
	    tmaddr += 2;
	    continue;
	}

	tile = videoRAM[tmaddr] | ((videoRAM[tmaddr+1] & 1) << 8);

	/* sprite palette? */
	palor = (videoRAM[tmaddr+1] & 8) << 1;

	/* flip vertical? */
	if (videoRAM[tmaddr+1] & 4) {
	    trow = ((tile << 6) | (urwt << 3));
	}
	else {
	    trow = ((tile << 6) | (rwt << 3));
	}

	/* flip horizontal? */
	if (!(videoRAM[tmaddr+1] & 2)) {
	    c = tileBitmaps[trow];
	    if (c) dest[(hs&255)] = paletteConverted[c | palor];
	    c = tileBitmaps[trow+1];
	    if (c) dest[(hs+1)&255] = paletteConverted[c | palor];
	    c = tileBitmaps[trow+2];
	    if (c) dest[(hs+2)&255] = paletteConverted[c | palor];
	    c = tileBitmaps[trow+3];
	    if (c) dest[(hs+3)&255] = paletteConverted[c | palor];
	    c = tileBitmaps[trow+4];
	    if (c) dest[(hs+4)&255] = paletteConverted[c | palor];
	    c = tileBitmaps[trow+5];
	    if (c) dest[(hs+5)&255] = paletteConverted[c | palor];
	    c = tileBitmaps[trow+6];
	    if (c) dest[(hs+6)&255] = paletteConverted[c | palor];
	    c = tileBitmaps[trow+7];
	    if (c) dest[(hs+7)&255] = paletteConverted[c | palor];
	}
	else {
	    c = tileBitmaps[trow];
	    if (c) dest[(hs+7)&255] = paletteConverted[c | palor];
	    c = tileBitmaps[trow+1];
	    if (c) dest[(hs+6)&255] = paletteConverted[c | palor];
	    c = tileBitmaps[trow+2];
	    if (c) dest[(hs+5)&255] = paletteConverted[c | palor];
	    c = tileBitmaps[trow+3];
	    if (c) dest[(hs+4)&255] = paletteConverted[c | palor];
	    c = tileBitmaps[trow+4];
	    if (c) dest[(hs+3)&255] = paletteConverted[c | palor];
	    c = tileBitmaps[trow+5];
	    if (c) dest[(hs+2)&255] = paletteConverted[c | palor];
	    c = tileBitmaps[trow+6];
	    if (c) dest[(hs+1)&255] = paletteConverted[c | palor];
	    c = tileBitmaps[trow+7];
	    if (c) dest[hs&255] = paletteConverted[c | palor];
	}
	hs += 8;
	tmaddr += 2;
    }
}

static void testSpriteCollisions()
{
    int i, j;
    int staddr;
    int col = 0;
    int sw, sh;
    int sx1, sx2, sy1, sy2;

    sw = 8;
    sh = 8;
    if (screenState.regs[1] & 1) {
	sw = 16;
	sh = 16;
    }
    if (screenState.regs[1] & 2) {
	sh = 16;
    }

    staddr = (screenState.regs[5] & 0x7e) << 7;
    for (i = 0; i < 64; i++) {
	sx1 = videoRAM[staddr+128+(i<<1)];
	sy1 = videoRAM[staddr+i];
	if (sy1 == 208) break;
	for (j = i+1; j < 64; j++) {
	    sx2 = videoRAM[staddr+128+(i<<1)];
	    sy2 = videoRAM[staddr+i];
	    if (sy2 == 208) break;
	    if (!((sx1 > (sx2+sw)) || (sx2 > (sx1+sw)) ||
		  (sy1 > (sy2+sh)) || (sy2 > (sy1+sh)))) {
		col = 1;
	    }
	}
    }

    if (col) {
	screenState.videoStatus |= 0x20;
    }
    else {
	screenState.videoStatus &= 0xdf;
    }
}

void MS_getScreenPixels(unsigned int *screen)
{
    unsigned int *s, *p, *src;
    int i, j, k;
    unsigned int pcol, c;

    int tmaddr;  /* tilemap address in VRAM */
    unsigned int tile;
    unsigned int trow;
    
    /* don't render if screen disabled */
    if (screenState.regs[1] & 0x40) {
	for (i = 0; i < 192; i++) {
	    memcpy(&screen[i*256], &screenBitmap[i<<8], 1024);
	}
    }

    testSpriteCollisions();
}
